@@ -1,42 +1,31 @@
 
 /*---------------------------------------------------------------*/
-/*---                                                         ---*/
-/*--- This file (libvex_guest_ppc32.h) is                     ---*/
-/*--- Copyright (C) OpenWorks LLP.  All rights reserved.      ---*/
-/*---                                                         ---*/
+/*--- begin                              libvex_guest_ppc32.h ---*/
 /*---------------------------------------------------------------*/
 
 /*
-   This file is part of LibVEX, a library for dynamic binary
-   instrumentation and translation.
+   This file is part of Valgrind, a dynamic binary instrumentation
+   framework.
 
-   Copyright (C) 2004-2007 OpenWorks LLP.  All rights reserved.
+   Copyright (C) 2004-2010 OpenWorks LLP
+      info@open-works.net
 
-   This library is made available under a dual licensing scheme.
+   This program is free software; you can redistribute it and/or
+   modify it under the terms of the GNU General Public License as
+   published by the Free Software Foundation; either version 2 of the
+   License, or (at your option) any later version.
 
-   If you link LibVEX against other code all of which is itself
-   licensed under the GNU General Public License, version 2 dated June
-   1991 ("GPL v2"), then you may use LibVEX under the terms of the GPL
-   v2, as appearing in the file LICENSE.GPL.  If the file LICENSE.GPL
-   is missing, you can obtain a copy of the GPL v2 from the Free
-   Software Foundation Inc., 51 Franklin St, Fifth Floor, Boston, MA
+   This program is distributed in the hope that it will be useful, but
+   WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   General Public License for more details.
+
+   You should have received a copy of the GNU General Public License
+   along with this program; if not, write to the Free Software
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301, USA.
 
-   For any other uses of LibVEX, you must first obtain a commercial
-   license from OpenWorks LLP.  Please contact info@open-works.co.uk
-   for information about commercial licensing.
-
-   This software is provided by OpenWorks LLP "as is" and any express
-   or implied warranties, including, but not limited to, the implied
-   warranties of merchantability and fitness for a particular purpose
-   are disclaimed.  In no event shall OpenWorks LLP be liable for any
-   direct, indirect, incidental, special, exemplary, or consequential
-   damages (including, but not limited to, procurement of substitute
-   goods or services; loss of use, data, or profits; or business
-   interruption) however caused and on any theory of liability,
-   whether in contract, strict liability, or tort (including
-   negligence or otherwise) arising in any way out of the use of this
-   software, even if advised of the possibility of such damage.
+   The GNU General Public License is contained in the file COPYING.
 
    Neither the names of the U.S. Department of Energy nor the
    University of California nor the names of its contributors may be
@@ -128,38 +117,41 @@ typedef
       /* 376 */ ULong guest_FPR31;
 
       // Vector Registers
-      /* 384 */ U128 guest_VR0 __attribute__ ((aligned (16)));
-      /* 400 */ U128 guest_VR1 __attribute__ ((aligned (16)));
-      /* 416 */ U128 guest_VR2 __attribute__ ((aligned (16)));
-      /* 432 */ U128 guest_VR3 __attribute__ ((aligned (16)));
-      /* 448 */ U128 guest_VR4 __attribute__ ((aligned (16)));
-      /* 464 */ U128 guest_VR5 __attribute__ ((aligned (16)));
-      /* 480 */ U128 guest_VR6 __attribute__ ((aligned (16)));
-      /* 496 */ U128 guest_VR7 __attribute__ ((aligned (16)));
-      /* 512 */ U128 guest_VR8 __attribute__ ((aligned (16)));
-      /* 528 */ U128 guest_VR9 __attribute__ ((aligned (16)));
-      /* 544 */ U128 guest_VR10 __attribute__ ((aligned (16)));
-      /* 560 */ U128 guest_VR11 __attribute__ ((aligned (16)));
-      /* 576 */ U128 guest_VR12 __attribute__ ((aligned (16)));
-      /* 592 */ U128 guest_VR13 __attribute__ ((aligned (16)));
-      /* 608 */ U128 guest_VR14 __attribute__ ((aligned (16)));
-      /* 624 */ U128 guest_VR15 __attribute__ ((aligned (16)));
-      /* 640 */ U128 guest_VR16 __attribute__ ((aligned (16)));
-      /* 656 */ U128 guest_VR17 __attribute__ ((aligned (16)));
-      /* 672 */ U128 guest_VR18 __attribute__ ((aligned (16)));
-      /* 688 */ U128 guest_VR19 __attribute__ ((aligned (16)));
-      /* 704 */ U128 guest_VR20 __attribute__ ((aligned (16)));
-      /* 720 */ U128 guest_VR21 __attribute__ ((aligned (16)));
-      /* 736 */ U128 guest_VR22 __attribute__ ((aligned (16)));
-      /* 752 */ U128 guest_VR23 __attribute__ ((aligned (16)));
-      /* 768 */ U128 guest_VR24 __attribute__ ((aligned (16)));
-      /* 784 */ U128 guest_VR25 __attribute__ ((aligned (16)));
-      /* 800 */ U128 guest_VR26 __attribute__ ((aligned (16)));
-      /* 816 */ U128 guest_VR27 __attribute__ ((aligned (16)));
-      /* 832 */ U128 guest_VR28 __attribute__ ((aligned (16)));
-      /* 848 */ U128 guest_VR29 __attribute__ ((aligned (16)));
-      /* 864 */ U128 guest_VR30 __attribute__ ((aligned (16)));
-      /* 880 */ U128 guest_VR31 __attribute__ ((aligned (16)));
+      // IMPORTANT: the user of libvex must place the guest state so as
+      // to ensure that guest_VR{0..31}, and any shadows thereof, are
+      // 16-aligned.
+      /* 384 */ U128 guest_VR0;
+      /* 400 */ U128 guest_VR1;
+      /* 416 */ U128 guest_VR2;
+      /* 432 */ U128 guest_VR3;
+      /* 448 */ U128 guest_VR4;
+      /* 464 */ U128 guest_VR5;
+      /* 480 */ U128 guest_VR6;
+      /* 496 */ U128 guest_VR7;
+      /* 512 */ U128 guest_VR8;
+      /* 528 */ U128 guest_VR9;
+      /* 544 */ U128 guest_VR10;
+      /* 560 */ U128 guest_VR11;
+      /* 576 */ U128 guest_VR12;
+      /* 592 */ U128 guest_VR13;
+      /* 608 */ U128 guest_VR14;
+      /* 624 */ U128 guest_VR15;
+      /* 640 */ U128 guest_VR16;
+      /* 656 */ U128 guest_VR17;
+      /* 672 */ U128 guest_VR18;
+      /* 688 */ U128 guest_VR19;
+      /* 704 */ U128 guest_VR20;
+      /* 720 */ U128 guest_VR21;
+      /* 736 */ U128 guest_VR22;
+      /* 752 */ U128 guest_VR23;
+      /* 768 */ U128 guest_VR24;
+      /* 784 */ U128 guest_VR25;
+      /* 800 */ U128 guest_VR26;
+      /* 816 */ U128 guest_VR27;
+      /* 832 */ U128 guest_VR28;
+      /* 848 */ U128 guest_VR29;
+      /* 864 */ U128 guest_VR30;
+      /* 880 */ U128 guest_VR31;
 
       /* 896 */ UInt guest_CIA;    // IP (no arch visible register)
       /* 900 */ UInt guest_LR;     // Link Register
@@ -201,13 +193,9 @@ typedef
       /* Emulation warnings */
       /* 940 */ UInt guest_EMWARN;
 
-      /* For lwarx/stwcx.: 0 == no reservation exists, non-0 == a
-         reservation exists. */
-      /* 944 */ UInt guest_RESVN;
-
       /* For icbi: record start and length of area to invalidate */
-      /* 948 */ UInt guest_TISTART;
-      /* 952 */ UInt guest_TILEN;
+      /* 944 */ UInt guest_TISTART;
+      /* 948 */ UInt guest_TILEN;
 
       /* Used to record the unredirected guest address at the start of
          a translation whose start has been redirected.  By reading
@@ -215,19 +203,20 @@ typedef
          find out what the corresponding no-redirection address was.
          Note, this is only set for wrap-style redirects, not for
          replace-style ones. */
-      /* 956 */ UInt guest_NRADDR;
-      /* 960 */ UInt guest_NRADDR_GPR2; /* needed by aix */
+      /* 952 */ UInt guest_NRADDR;
+      /* 956 */ UInt guest_NRADDR_GPR2; /* needed by aix */
 
      /* A grows-upwards stack for hidden saves/restores of LR and R2
         needed for function interception and wrapping on ppc32-aix5.
         A horrible hack.  REDIR_SP points to the highest live entry,
         and so starts at -1. */
-      /* 964 */ UInt guest_REDIR_SP;
-      /* 968 */ UInt guest_REDIR_STACK[VEX_GUEST_PPC32_REDIR_STACK_SIZE];
+      /* 960 */ UInt guest_REDIR_SP;
+      /* 964 */ UInt guest_REDIR_STACK[VEX_GUEST_PPC32_REDIR_STACK_SIZE];
 
-      /* Needed for AIX: CIA at the last SC insn.  Used when backing up
-         to restart a syscall that has been interrupted by a signal. */
-      /* ??? */ UInt guest_CIA_AT_SC; 
+      /* Needed for AIX (but mandated for all guest architectures):
+         CIA at the last SC insn.  Used when backing up to restart a
+         syscall that has been interrupted by a signal. */
+      /* ??? */ UInt guest_IP_AT_SYSCALL; 
 
       /* SPRG3, which AIUI is readonly in user space.  Needed for
          threading on AIX. */
