@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Julian Seward 
+   Copyright (C) 2000-2010 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -41,6 +41,11 @@
 #include "pub_core_clientstate.h"   // VG_(fd_hard_limit)
 #include "pub_core_syscall.h"
 
+/* IMPORTANT: on Darwin it is essential to use the _nocancel versions
+   of syscalls rather than the vanilla version, if a _nocancel version
+   is available.  See docs/internals/Darwin-notes.txt for the reason
+   why. */
+
 /* ---------------------------------------------------------------------
    File stuff
    ------------------------------------------------------------------ */
@@ -74,37 +79,76 @@ Int VG_(safe_fd)(Int oldfd)
    or if it doesn't exist, we return False. */
 Bool VG_(resolve_filename) ( Int fd, HChar* buf, Int n_buf )
 {
+#  if defined(VGO_linux)
    HChar tmp[64];
-
    VG_(sprintf)(tmp, "/proc/self/fd/%d", fd);
    VG_(memset)(buf, 0, n_buf);
-
    if (VG_(readlink)(tmp, buf, n_buf) > 0 && buf[0] == '/')
       return True;
    else
       return False;
+
+#  elif defined(VGO_aix5)
+   I_die_here; /* maybe just return False? */
+   return False;
+
+#  elif defined(VGO_darwin)
+   HChar tmp[VKI_MAXPATHLEN+1];
+   if (0 == VG_(fcntl)(fd, VKI_F_GETPATH, (UWord)tmp)) {
+      if (n_buf > 0) {
+         VG_(strncpy)( buf, tmp, n_buf < sizeof(tmp) ? n_buf : sizeof(tmp) );
+         buf[n_buf-1] = 0;
+      }
+      if (tmp[0] == '/') return True;
+   }
+   return False;
+
+#  else
+#     error Unknown OS
+#  endif
 }
 
 SysRes VG_(open) ( const Char* pathname, Int flags, Int mode )
 {  
-   SysRes res = VG_(do_syscall3)(__NR_open, (UWord)pathname, flags, mode);
+#  if defined(VGO_linux) || defined(VGO_aix5)
+   SysRes res = VG_(do_syscall3)(__NR_open,
+                                 (UWord)pathname, flags, mode);
+#  elif defined(VGO_darwin)
+   SysRes res = VG_(do_syscall3)(__NR_open_nocancel,
+                                 (UWord)pathname, flags, mode);
+#  else
+#    error Unknown OS
+#  endif
    return res;
 }
 
 void VG_(close) ( Int fd )
 {
+   /* Hmm.  Return value is not checked.  That's uncool. */
+#  if defined(VGO_linux) || defined(VGO_aix5)
    (void)VG_(do_syscall1)(__NR_close, fd);
+#  elif defined(VGO_darwin)
+   (void)VG_(do_syscall1)(__NR_close_nocancel, fd);
+#  else
+#    error Unknown OS
+#  endif
 }
 
 Int VG_(read) ( Int fd, void* buf, Int count)
 {
    Int    ret;
+#  if defined(VGO_linux) || defined(VGO_aix5)
    SysRes res = VG_(do_syscall3)(__NR_read, fd, (UWord)buf, count);
-   if (res.isError) {
-      ret = - (Int)(Word)res.err;
+#  elif defined(VGO_darwin)
+   SysRes res = VG_(do_syscall3)(__NR_read_nocancel, fd, (UWord)buf, count);
+#  else
+#    error Unknown OS
+#  endif
+   if (sr_isError(res)) {
+      ret = - (Int)(Word)sr_Err(res);
       vg_assert(ret < 0);
    } else {
-      ret = (Int)(Word)res.res;
+      ret = (Int)(Word)sr_Res(res);
       vg_assert(ret >= 0);
    }
    return ret;
@@ -113,27 +157,54 @@ Int VG_(read) ( Int fd, void* buf, Int count)
 Int VG_(write) ( Int fd, const void* buf, Int count)
 {
    Int    ret;
+#  if defined(VGO_linux) || defined(VGO_aix5)
    SysRes res = VG_(do_syscall3)(__NR_write, fd, (UWord)buf, count);
-   if (res.isError) {
-      ret = - (Int)(Word)res.err;
+#  elif defined(VGO_darwin)
+   SysRes res = VG_(do_syscall3)(__NR_write_nocancel, fd, (UWord)buf, count);
+#  else
+#    error "Unknown OS"
+#  endif
+   if (sr_isError(res)) {
+      ret = - (Int)(Word)sr_Err(res);
       vg_assert(ret < 0);
    } else {
-      ret = (Int)(Word)res.res;
+      ret = (Int)(Word)sr_Res(res);
       vg_assert(ret >= 0);
    }
    return ret;
 }
 
+
 Int VG_(pipe) ( Int fd[2] )
 {
+#  if defined(VGO_linux) || defined(VGO_aix5)
    SysRes res = VG_(do_syscall1)(__NR_pipe, (UWord)fd);
-   return res.isError ? -1 : 0;
+   return sr_isError(res) ? -1 : 0;
+#  elif defined(VGO_darwin)
+   /* __NR_pipe is UX64, so produces a double-word result */
+   SysRes res = VG_(do_syscall0)(__NR_pipe);
+   if (!sr_isError(res)) {
+      fd[0] = (Int)sr_Res(res);
+      fd[1] = (Int)sr_ResHI(res);
+   }
+   return sr_isError(res) ? -1 : 0;
+#  else
+#    error "Unknown OS"
+#  endif
 }
 
 OffT VG_(lseek) ( Int fd, OffT offset, Int whence )
 {
+#  if defined(VGO_linux) || defined(VGO_aix5) || defined(VGP_amd64_darwin)
    SysRes res = VG_(do_syscall3)(__NR_lseek, fd, offset, whence);
-   return res.isError ? (-1) : res.res;
+   vg_assert(sizeof(OffT) == sizeof(Word));
+#  elif defined(VGP_x86_darwin)
+   SysRes res = VG_(do_syscall4)(__NR_lseek, fd, 
+                                 offset & 0xffffffff, offset >> 32, whence);
+#  else
+#    error "Unknown plat"
+#  endif
+   return sr_isError(res) ? (-1) : sr_Res(res);
    /* if you change the error-reporting conventions of this, also
       change VG_(pread) and all other usage points. */
 }
@@ -145,46 +216,50 @@ OffT VG_(lseek) ( Int fd, OffT offset, Int whence )
 
 #define TRANSLATE_TO_vg_stat(_p_vgstat, _p_vkistat) \
    do { \
-      (_p_vgstat)->st_dev        = (ULong)( (_p_vkistat)->st_dev ); \
-      (_p_vgstat)->st_ino        = (ULong)( (_p_vkistat)->st_ino ); \
-      (_p_vgstat)->st_nlink      = (ULong)( (_p_vkistat)->st_nlink ); \
-      (_p_vgstat)->st_mode       = (UInt)( (_p_vkistat)->st_mode ); \
-      (_p_vgstat)->st_uid        = (UInt)( (_p_vkistat)->st_uid ); \
-      (_p_vgstat)->st_gid        = (UInt)( (_p_vkistat)->st_gid ); \
-      (_p_vgstat)->st_rdev       = (ULong)( (_p_vkistat)->st_rdev ); \
-      (_p_vgstat)->st_size       = (Long)( (_p_vkistat)->st_size ); \
-      (_p_vgstat)->st_blksize    = (ULong)( (_p_vkistat)->st_blksize ); \
-      (_p_vgstat)->st_blocks     = (ULong)( (_p_vkistat)->st_blocks ); \
-      (_p_vgstat)->st_atime      = (ULong)( (_p_vkistat)->st_atime ); \
-      (_p_vgstat)->st_atime_nsec = (ULong)( (_p_vkistat)->st_atime_nsec ); \
-      (_p_vgstat)->st_mtime      = (ULong)( (_p_vkistat)->st_mtime ); \
-      (_p_vgstat)->st_mtime_nsec = (ULong)( (_p_vkistat)->st_mtime_nsec ); \
-      (_p_vgstat)->st_ctime      = (ULong)( (_p_vkistat)->st_ctime ); \
-      (_p_vgstat)->st_ctime_nsec = (ULong)( (_p_vkistat)->st_ctime_nsec ); \
+      (_p_vgstat)->dev        = (ULong)( (_p_vkistat)->st_dev ); \
+      (_p_vgstat)->ino        = (ULong)( (_p_vkistat)->st_ino ); \
+      (_p_vgstat)->nlink      = (ULong)( (_p_vkistat)->st_nlink ); \
+      (_p_vgstat)->mode       = (UInt) ( (_p_vkistat)->st_mode ); \
+      (_p_vgstat)->uid        = (UInt) ( (_p_vkistat)->st_uid ); \
+      (_p_vgstat)->gid        = (UInt) ( (_p_vkistat)->st_gid ); \
+      (_p_vgstat)->rdev       = (ULong)( (_p_vkistat)->st_rdev ); \
+      (_p_vgstat)->size       = (Long) ( (_p_vkistat)->st_size ); \
+      (_p_vgstat)->blksize    = (ULong)( (_p_vkistat)->st_blksize ); \
+      (_p_vgstat)->blocks     = (ULong)( (_p_vkistat)->st_blocks ); \
+      (_p_vgstat)->atime      = (ULong)( (_p_vkistat)->st_atime ); \
+      (_p_vgstat)->atime_nsec = (ULong)( (_p_vkistat)->st_atime_nsec ); \
+      (_p_vgstat)->mtime      = (ULong)( (_p_vkistat)->st_mtime ); \
+      (_p_vgstat)->mtime_nsec = (ULong)( (_p_vkistat)->st_mtime_nsec ); \
+      (_p_vgstat)->ctime      = (ULong)( (_p_vkistat)->st_ctime ); \
+      (_p_vgstat)->ctime_nsec = (ULong)( (_p_vkistat)->st_ctime_nsec ); \
    } while (0)
 
-SysRes VG_(stat) ( Char* file_name, struct vg_stat* vgbuf )
+SysRes VG_(stat) ( const Char* file_name, struct vg_stat* vgbuf )
 {
    SysRes res;
    VG_(memset)(vgbuf, 0, sizeof(*vgbuf));
-#  if defined(VGO_linux)
+
+#  if defined(VGO_linux) || defined(VGO_darwin)
+   /* First try with stat64.  If that doesn't work out, fall back to
+      the vanilla version. */
 #  if defined(__NR_stat64)
    { struct vki_stat64 buf64;
      res = VG_(do_syscall2)(__NR_stat64, (UWord)file_name, (UWord)&buf64);
-     if (!(res.isError && res.err == VKI_ENOSYS)) {
+     if (!(sr_isError(res) && sr_Err(res) == VKI_ENOSYS)) {
         /* Success, or any failure except ENOSYS */
-        if (!res.isError)
+        if (!sr_isError(res))
            TRANSLATE_TO_vg_stat(vgbuf, &buf64);
         return res;
      }
    }
-#  endif /* if defined(__NR_stat64) */
+#  endif /* defined(__NR_stat64) */
    { struct vki_stat buf;
      res = VG_(do_syscall2)(__NR_stat, (UWord)file_name, (UWord)&buf);
-     if (!res.isError)
+     if (!sr_isError(res))
         TRANSLATE_TO_vg_stat(vgbuf, &buf);
      return res;
    }
+
 #  elif defined(VGO_aix5)
    { struct vki_stat buf;
      res = VG_(do_syscall4)(__NR_AIX5_statx,
@@ -192,17 +267,18 @@ SysRes VG_(stat) ( Char* file_name, struct vg_stat* vgbuf )
                             (UWord)&buf,
                             sizeof(struct vki_stat),
                             VKI_STX_NORMAL);
-     if (!res.isError) {
+     if (!sr_isError(res)) {
         VG_(memset)(vgbuf, 0, sizeof(*vgbuf));
-        vgbuf->st_dev  = (ULong)buf.st_dev;
-        vgbuf->st_ino  = (ULong)buf.st_ino;
-        vgbuf->st_mode = (UInt)buf.st_mode;
-        vgbuf->st_uid  = (UInt)buf.st_uid;
-        vgbuf->st_gid  = (UInt)buf.st_gid;
-        vgbuf->st_size = (Long)buf.st_size;
+        vgbuf->dev  = (ULong)buf.st_dev;
+        vgbuf->ino  = (ULong)buf.st_ino;
+        vgbuf->mode = (UInt)buf.st_mode;
+        vgbuf->uid  = (UInt)buf.st_uid;
+        vgbuf->gid  = (UInt)buf.st_gid;
+        vgbuf->size = (Long)buf.st_size;
      }
      return res;
    }
+
 #  else
 #    error Unknown OS
 #  endif
@@ -212,26 +288,31 @@ Int VG_(fstat) ( Int fd, struct vg_stat* vgbuf )
 {
    SysRes res;
    VG_(memset)(vgbuf, 0, sizeof(*vgbuf));
-#  if defined(VGO_linux)
+
+#  if defined(VGO_linux)  ||  defined(VGO_darwin)
+   /* First try with fstat64.  If that doesn't work out, fall back to
+      the vanilla version. */
 #  if defined(__NR_fstat64)
    { struct vki_stat64 buf64;
      res = VG_(do_syscall2)(__NR_fstat64, (UWord)fd, (UWord)&buf64);
-     if (!(res.isError && res.err == VKI_ENOSYS)) {
+     if (!(sr_isError(res) && sr_Err(res) == VKI_ENOSYS)) {
         /* Success, or any failure except ENOSYS */
-        if (!res.isError)
+        if (!sr_isError(res))
            TRANSLATE_TO_vg_stat(vgbuf, &buf64);
-        return res.isError ? (-1) : 0;
+        return sr_isError(res) ? (-1) : 0;
      }
    }
 #  endif /* if defined(__NR_fstat64) */
    { struct vki_stat buf;
      res = VG_(do_syscall2)(__NR_fstat, (UWord)fd, (UWord)&buf);
-     if (!res.isError)
+     if (!sr_isError(res))
         TRANSLATE_TO_vg_stat(vgbuf, &buf);
-     return res.isError ? (-1) : 0;
+     return sr_isError(res) ? (-1) : 0;
    }
+
 #  elif defined(VGO_aix5)
    I_die_here;
+
 #  else
 #    error Unknown OS
 #  endif
@@ -244,15 +325,15 @@ Long VG_(fsize) ( Int fd )
 {
    struct vg_stat buf;
    Int res = VG_(fstat)( fd, &buf );
-   return (res == -1) ? (-1LL) : buf.st_size;
+   return (res == -1) ? (-1LL) : buf.size;
 }
 
-Bool VG_(is_dir) ( HChar* f )
+Bool VG_(is_dir) ( const HChar* f )
 {
    struct vg_stat buf;
    SysRes res = VG_(stat)(f, &buf);
-   return res.isError ? False
-                      : VKI_S_ISDIR(buf.st_mode) ? True : False;
+   return sr_isError(res) ? False
+                      : VKI_S_ISDIR(buf.mode) ? True : False;
 }
 
 SysRes VG_(dup) ( Int oldfd )
@@ -262,7 +343,7 @@ SysRes VG_(dup) ( Int oldfd )
 
 SysRes VG_(dup2) ( Int oldfd, Int newfd )
 {
-#  if defined(VGO_linux)
+#  if defined(VGO_linux)  ||  defined(VGO_darwin)
    return VG_(do_syscall2)(__NR_dup2, oldfd, newfd);
 #  elif defined(VGO_aix5)
    I_die_here;
@@ -272,22 +353,28 @@ SysRes VG_(dup2) ( Int oldfd, Int newfd )
 }
 
 /* Returns -1 on error. */
-Int VG_(fcntl) ( Int fd, Int cmd, Int arg )
+Int VG_(fcntl) ( Int fd, Int cmd, Addr arg )
 {
+#  if defined(VGO_linux) || defined(VGO_aix5)
    SysRes res = VG_(do_syscall3)(__NR_fcntl, fd, cmd, arg);
-   return res.isError ? -1 : res.res;
+#  elif defined(VGO_darwin)
+   SysRes res = VG_(do_syscall3)(__NR_fcntl_nocancel, fd, cmd, arg);
+#  else
+#    error "Unknown OS"
+#  endif
+   return sr_isError(res) ? -1 : sr_Res(res);
 }
 
-Int VG_(rename) ( Char* old_name, Char* new_name )
+Int VG_(rename) ( const Char* old_name, const Char* new_name )
 {
    SysRes res = VG_(do_syscall2)(__NR_rename, (UWord)old_name, (UWord)new_name);
-   return res.isError ? (-1) : 0;
+   return sr_isError(res) ? (-1) : 0;
 }
 
-Int VG_(unlink) ( Char* file_name )
+Int VG_(unlink) ( const Char* file_name )
 {
    SysRes res = VG_(do_syscall1)(__NR_unlink, (UWord)file_name);
-   return res.isError ? (-1) : 0;
+   return sr_isError(res) ? (-1) : 0;
 }
 
 /* The working directory at startup.  AIX doesn't provide an easy
@@ -315,15 +402,15 @@ Bool VG_(record_startup_wd) ( void )
    { SysRes res
         = VG_(do_syscall2)(__NR_getcwd, (UWord)startup_wd, szB-1);
      vg_assert(startup_wd[szB-1] == 0);
-     if (res.isError) {
+     if (sr_isError(res)) {
         return False;
      } else {
         startup_wd_acquired = True;
         return True;
      }
    }
-#  elif defined(VGO_aix5)
-   /* We can't ask the kernel, so instead rely on launcher-aix5.c to
+#  elif defined(VGO_aix5) || defined(VGO_darwin)
+   /* We can't ask the kernel, so instead rely on launcher-*.c to
       tell us the startup path.  Note the env var is keyed to the
       parent's PID, not ours, since our parent is the launcher
       process. */
@@ -357,44 +444,52 @@ Bool VG_(get_startup_wd) ( Char* buf, SizeT size )
    return True;
 }
 
-Int VG_(readlink) (Char* path, Char* buf, UInt bufsiz)
+Int VG_(readlink) (const Char* path, Char* buf, UInt bufsiz)
 {
    SysRes res;
    /* res = readlink( path, buf, bufsiz ); */
    res = VG_(do_syscall3)(__NR_readlink, (UWord)path, (UWord)buf, bufsiz);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 }
 
-Int VG_(getdents) (UInt fd, struct vki_dirent *dirp, UInt count)
+Int VG_(getdents) (Int fd, struct vki_dirent *dirp, UInt count)
 {
+#  if defined(VGO_linux) || defined(VGO_aix5)
    SysRes res;
    /* res = getdents( fd, dirp, count ); */
    res = VG_(do_syscall3)(__NR_getdents, fd, (UWord)dirp, count);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
+#  elif defined(VGO_darwin)
+   I_die_here;
+#  else
+#    error "Unknown OS"
+#  endif
 }
 
 /* Check accessibility of a file.  Returns zero for access granted,
    nonzero otherwise. */
-Int VG_(access) ( HChar* path, Bool irusr, Bool iwusr, Bool ixusr )
+Int VG_(access) ( const HChar* path, Bool irusr, Bool iwusr, Bool ixusr )
 {
 #  if defined(VGO_linux)
    /* Very annoyingly, I cannot find any definition for R_OK et al in
       the kernel interfaces.  Therefore I reluctantly resort to
       hardwiring in these magic numbers that I determined by
       experimentation. */
-   UWord w = (irusr ? 4/*R_OK*/ : 0)
-             | (iwusr ? 2/*W_OK*/ : 0)
-             | (ixusr ? 1/*X_OK*/ : 0);
-   SysRes res = VG_(do_syscall2)(__NR_access, (UWord)path, w);
-   return res.isError ? 1 : 0;
-#  elif defined(VGO_aix5)
+#  define VKI_R_OK 4
+#  define VKI_W_OK 2
+#  define VKI_X_OK 1
+#  endif
+
    UWord w = (irusr ? VKI_R_OK : 0)
              | (iwusr ? VKI_W_OK : 0)
              | (ixusr ? VKI_X_OK : 0);
    SysRes res = VG_(do_syscall2)(__NR_access, (UWord)path, w);
-   return res.isError ? 1 : 0;   
-#  else
-#    error "Don't know how to do VG_(access) on this OS"
+   return sr_isError(res) ? 1 : 0;   
+
+#  if defined(VGO_linux)
+#  undef VKI_R_OK
+#  undef VKI_W_OK
+#  undef VKI_X_OK
 #  endif
 }
 
@@ -419,7 +514,7 @@ Int VG_(access) ( HChar* path, Bool irusr, Bool iwusr, Bool ixusr )
 */
 /* returns: 0 = success, non-0 is failure */
 Int VG_(check_executable)(/*OUT*/Bool* is_setuid,
-                          HChar* f, Bool allow_setuid)
+                          const HChar* f, Bool allow_setuid)
 {
    struct vg_stat st;
    SysRes res = VG_(stat)(f, &st);
@@ -427,23 +522,23 @@ Int VG_(check_executable)(/*OUT*/Bool* is_setuid,
    if (is_setuid)
       *is_setuid = False;
 
-   if (res.isError) {
-      return res.err;
+   if (sr_isError(res)) {
+      return sr_Err(res);
    }
 
-   if ( (st.st_mode & (VKI_S_ISUID | VKI_S_ISGID)) && !allow_setuid ) {
+   if ( (st.mode & (VKI_S_ISUID | VKI_S_ISGID)) && !allow_setuid ) {
       if (is_setuid)
          *is_setuid = True;
       return VKI_EACCES;
    }
 
-   if (VG_(geteuid)() == st.st_uid) {
-      if (!(st.st_mode & VKI_S_IXUSR))
+   if (VG_(geteuid)() == st.uid) {
+      if (!(st.mode & VKI_S_IXUSR))
          return VKI_EACCES;
    } else {
       Int grpmatch = 0;
 
-      if (VG_(getegid)() == st.st_gid)
+      if (VG_(getegid)() == st.gid)
 	 grpmatch = 1;
       else {
 	 UInt groups[32];
@@ -451,7 +546,7 @@ Int VG_(check_executable)(/*OUT*/Bool* is_setuid,
 	 Int i;
          /* ngrp will be -1 if VG_(getgroups) failed. */
          for (i = 0; i < ngrp; i++) {
-	    if (groups[i] == st.st_gid) {
+	    if (groups[i] == st.gid) {
 	       grpmatch = 1;
 	       break;
 	    }
@@ -459,10 +554,10 @@ Int VG_(check_executable)(/*OUT*/Bool* is_setuid,
       }
 
       if (grpmatch) {
-	 if (!(st.st_mode & VKI_S_IXGRP)) {
+	 if (!(st.mode & VKI_S_IXGRP)) {
             return VKI_EACCES;
          }
-      } else if (!(st.st_mode & VKI_S_IXOTH)) {
+      } else if (!(st.mode & VKI_S_IXOTH)) {
          return VKI_EACCES;
       }
    }
@@ -470,12 +565,31 @@ Int VG_(check_executable)(/*OUT*/Bool* is_setuid,
    return 0;
 }
 
-SysRes VG_(pread) ( Int fd, void* buf, Int count, Int offset )
+/* DDD: Note this moves (or at least, is believed to move) the file pointer
+   on Linux and AIX5 but doesn't on Darwin.  This inconsistency should
+   be fixed.  (In other words, why isn't the Linux/AIX5 version implemented in
+   terms of pread()?) */
+SysRes VG_(pread) ( Int fd, void* buf, Int count, OffT offset )
 {
-   OffT off = VG_(lseek)( fd, (OffT)offset, VKI_SEEK_SET);
+   SysRes res;
+#  if defined(VGO_linux) || defined(VGO_aix5)
+   /* Linux, AIX5 */
+   OffT off = VG_(lseek)( fd, offset, VKI_SEEK_SET);
    if (off < 0)
       return VG_(mk_SysRes_Error)( VKI_EINVAL );
-   return VG_(do_syscall3)(__NR_read, fd, (UWord)buf, count );
+   res = VG_(do_syscall3)(__NR_read, fd, (UWord)buf, count );
+   return res;
+#  elif defined(VGP_amd64_darwin)
+   res = VG_(do_syscall4)(__NR_pread_nocancel, fd, (UWord)buf, count, offset);
+   return res;
+#  elif defined(VGP_x86_darwin)
+   /* ppc32-darwin is the same, but with the args inverted */
+   res = VG_(do_syscall5)(__NR_pread_nocancel, fd, (UWord)buf, count, 
+                          offset & 0xffffffff, offset >> 32);
+   return res;
+#  else
+#    error "Unknown platform"
+#  endif
 }
 
 /* Create and open (-rw------) a tmp file name incorporating said arg.
@@ -508,10 +622,10 @@ Int VG_(mkstemp) ( HChar* part_of_name, /*OUT*/HChar* fullname )
       sres = VG_(open)(buf,
                        VKI_O_CREAT|VKI_O_RDWR|VKI_O_EXCL|VKI_O_TRUNC,
                        VKI_S_IRUSR|VKI_S_IWUSR);
-      if (sres.isError)
+      if (sr_isError(sres))
          continue;
       /* VG_(safe_fd) doesn't return if it fails. */
-      fd = VG_(safe_fd)( sres.res );
+      fd = VG_(safe_fd)( sr_Res(sres) );
       if (fullname)
          VG_(strcpy)( fullname, buf );
       return fd;
@@ -521,23 +635,18 @@ Int VG_(mkstemp) ( HChar* part_of_name, /*OUT*/HChar* fullname )
 
 
 /* ---------------------------------------------------------------------
-   Socket-related stuff.  This is very Linux-kernel specific.
+   Socket-related stuff.
    ------------------------------------------------------------------ */
 
+#if defined(VGO_aix5)
+struct vki_sockaddr_in;
+#endif
+
 static
 Int parse_inet_addr_and_port ( UChar* str, UInt* ip_addr, UShort* port );
 
 static
-Int my_socket ( Int domain, Int type, Int protocol );
-
-static
-Int my_connect ( Int sockfd, 
-#                if defined(VGO_linux)
-                   struct vki_sockaddr_in* serv_addr, 
-#                else
-                   void* serv_addr,
-#                endif
-                 Int addrlen );
+Int my_connect ( Int sockfd, struct vki_sockaddr_in* serv_addr, Int addrlen );
 
 UInt VG_(htonl) ( UInt x )
 {
@@ -594,9 +703,7 @@ UShort VG_(ntohs) ( UShort x )
 */
 Int VG_(connect_via_socket)( UChar* str )
 {
-#if defined(VGO_aix5)
-   I_die_here;
-#else /* Yay, Linux */
+#  if defined(VGO_linux) || defined(VGO_darwin)
    Int sd, res;
    struct vki_sockaddr_in servAddr;
    UInt   ip   = 0;
@@ -616,22 +723,27 @@ Int VG_(connect_via_socket)( UChar* str )
    servAddr.sin_port = VG_(htons)(port);
 
    /* create socket */
-   sd = my_socket(VKI_AF_INET, VKI_SOCK_STREAM, 0 /* IPPROTO_IP ? */);
+   sd = VG_(socket)(VKI_AF_INET, VKI_SOCK_STREAM, 0 /* IPPROTO_IP ? */);
    if (sd < 0) {
       /* this shouldn't happen ... nevertheless */
       return -2;
    }
 		
    /* connect to server */
-   res = my_connect(sd, (struct vki_sockaddr_in *) &servAddr, 
-                        sizeof(servAddr));
+   res = my_connect(sd, &servAddr, sizeof(servAddr));
    if (res < 0) {
       /* connection failed */
       return -2;
    }
 
    return sd;
-#endif
+
+#  elif defined(VGO_aix5)
+   I_die_here;
+
+#  else
+#    error "Unknown OS"
+#  endif
 }
 
 
@@ -679,70 +791,92 @@ static Int parse_inet_addr_and_port ( UChar* str, UInt* ip_addr, UShort* port )
 #  undef GET_CH
 }
 
-static
-Int my_socket ( Int domain, Int type, Int protocol )
+// GrP fixme safe_fd?
+Int VG_(socket) ( Int domain, Int type, Int protocol )
 {
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[3];
    args[0] = domain;
    args[1] = type;
    args[2] = protocol;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_SOCKET, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall3)(__NR_socket, domain, type, protocol );
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGO_darwin)
+   SysRes res;
+   res = VG_(do_syscall3)(__NR_socket, domain, type, protocol);
+   if (!sr_isError(res)) {
+       // Set SO_NOSIGPIPE so write() returns EPIPE instead of raising SIGPIPE
+       Int optval = 1;
+       SysRes res2;
+       res2 = VG_(do_syscall5)(__NR_setsockopt, sr_Res(res), VKI_SOL_SOCKET, 
+                               VKI_SO_NOSIGPIPE, (UWord)&optval, 
+                               sizeof(optval));
+       // ignore setsockopt() error
+   }
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown arch"
+#  endif
 }
 
+
 static
-Int my_connect ( Int sockfd,
-#                if defined(VGO_linux)
-                   struct vki_sockaddr_in* serv_addr, 
-#                else
-                   void* serv_addr,
-#                endif
-                 Int addrlen )
+Int my_connect ( Int sockfd, struct vki_sockaddr_in* serv_addr, Int addrlen )
 {
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[3];
    args[0] = sockfd;
    args[1] = (UWord)serv_addr;
    args[2] = addrlen;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_CONNECT, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall3)(__NR_connect, sockfd, (UWord)serv_addr, addrlen);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGO_darwin)
+   SysRes res;
+   res = VG_(do_syscall3)(__NR_connect_nocancel,
+                          sockfd, (UWord)serv_addr, addrlen);
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown arch"
+#  endif
 }
 
 Int VG_(write_socket)( Int sd, void *msg, Int count )
 {
    /* This is actually send(). */
-   /* For Linux, VKI_MSG_NOSIGNAL is a request not to send SIGPIPE on
-      errors on stream oriented sockets when the other end breaks the
-      connection. The EPIPE error is still returned. */
 
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+   /* For Linux, VKI_MSG_NOSIGNAL is a request not to send SIGPIPE on 
+      errors on stream oriented sockets when the other end breaks the
+      connection. The EPIPE error is still returned.
+
+      For Darwin, VG_(socket)() sets SO_NOSIGPIPE to get EPIPE instead of 
+      SIGPIPE */
+
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[4];
    args[0] = sd;
@@ -750,76 +884,96 @@ Int VG_(write_socket)( Int sd, void *msg, Int count )
    args[2] = count;
    args[3] = VKI_MSG_NOSIGNAL;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_SEND, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall6)(__NR_sendto, sd, (UWord)msg, 
                                        count, VKI_MSG_NOSIGNAL, 0,0);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGP_x86_darwin) || defined(VGP_amd64_darwin)
+   SysRes res;
+   res = VG_(do_syscall3)(__NR_write_nocancel, sd, (UWord)msg, count);
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown platform"
+#  endif
 }
 
 Int VG_(getsockname) ( Int sd, struct vki_sockaddr *name, Int *namelen)
 {
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[3];
    args[0] = sd;
    args[1] = (UWord)name;
    args[2] = (UWord)namelen;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_GETSOCKNAME, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall3)( __NR_getsockname,
                            (UWord)sd, (UWord)name, (UWord)namelen );
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGO_darwin)
+   SysRes res;
+   res = VG_(do_syscall3)( __NR_getsockname,
+                           (UWord)sd, (UWord)name, (UWord)namelen );
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown platform"
+#  endif
 }
 
 Int VG_(getpeername) ( Int sd, struct vki_sockaddr *name, Int *namelen)
 {
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[3];
    args[0] = sd;
    args[1] = (UWord)name;
    args[2] = (UWord)namelen;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_GETPEERNAME, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall3)( __NR_getpeername,
                            (UWord)sd, (UWord)name, (UWord)namelen );
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGO_darwin)
+   SysRes res;
+   res = VG_(do_syscall3)( __NR_getpeername,
+                           (UWord)sd, (UWord)name, (UWord)namelen );
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown platform"
+#  endif
 }
 
 Int VG_(getsockopt) ( Int sd, Int level, Int optname, void *optval,
                       Int *optlen)
 {
-#if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux)
+#  if defined(VGP_x86_linux) || defined(VGP_ppc32_linux) \
+      || defined(VGP_ppc64_linux)
    SysRes res;
    UWord  args[5];
    args[0] = sd;
@@ -828,26 +982,108 @@ Int VG_(getsockopt) ( Int sd, Int level, Int optname, void *optval,
    args[3] = (UWord)optval;
    args[4] = (UWord)optlen;
    res = VG_(do_syscall2)(__NR_socketcall, VKI_SYS_GETSOCKOPT, (UWord)&args);
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_amd64_linux)
+#  elif defined(VGP_amd64_linux) || defined(VGP_arm_linux)
    SysRes res;
    res = VG_(do_syscall5)( __NR_getsockopt,
                            (UWord)sd, (UWord)level, (UWord)optname, 
                            (UWord)optval, (UWord)optlen );
-   return res.isError ? -1 : res.res;
+   return sr_isError(res) ? -1 : sr_Res(res);
 
-#elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#  elif defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
    I_die_here;
 
-#else
-#  error Unknown arch
-#endif
+#  elif defined(VGO_darwin)
+   SysRes res;
+   res = VG_(do_syscall5)( __NR_getsockopt,
+                           (UWord)sd, (UWord)level, (UWord)optname, 
+                           (UWord)optval, (UWord)optlen );
+   return sr_isError(res) ? -1 : sr_Res(res);
+
+#  else
+#    error "Unknown platform"
+#  endif
 }
 
 
+Char *VG_(basename)(const Char *path)
+{
+   static Char buf[VKI_PATH_MAX];
+   
+   const Char *p, *end;
+
+   if (path == NULL  ||  
+       0 == VG_(strcmp)(path, ""))
+   {
+      return ".";
+   }
+
+   p = path + VG_(strlen)(path);
+   while (p > path  &&  *p == '/') {
+      // skip all trailing '/'
+      p--;
+   }
+
+   if (p == path  &&  *p == '/') return "/"; // all slashes
+
+   end = p;
+
+   while (p > path  &&  *p != '/') {
+      // now skip non '/'
+      p--;
+   }
+
+   if (*p == '/') p++;
+
+   VG_(strncpy)(buf, p, end-p+1);
+   buf[end-p+1] = '\0';
+
+   return buf;
+}
+
+
+Char *VG_(dirname)(const Char *path)
+{
+   static Char buf[VKI_PATH_MAX];
+    
+   const Char *p;
+
+   if (path == NULL  ||  
+       0 == VG_(strcmp)(path, "")  ||  
+       0 == VG_(strcmp)(path, "/"))
+   {
+      return ".";
+   }
+
+   p = path + VG_(strlen)(path);
+   while (p > path  &&  *p == '/') {
+      // skip all trailing '/'
+      p--;
+   }
+
+   while (p > path  &&  *p != '/') {
+      // now skip non '/'
+      p--;
+   }
+
+   if (p == path) {
+      if (*p == '/') return "/"; // all slashes
+      else return "."; // no slashes
+   } 
+
+   while (p > path  &&  *p == '/') {
+      // skip '/' again
+      p--;
+   }
+
+   VG_(strncpy)(buf, path, p-path+1);
+   buf[p-path+1] = '\0';
+
+   return buf;
+}
+
 
 /*--------------------------------------------------------------------*/
 /*--- end                                                          ---*/
 /*--------------------------------------------------------------------*/
-
