@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Julian Seward 
+   Copyright (C) 2000-2010 Julian Seward 
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -31,7 +31,7 @@
 #include "pub_core_basics.h"
 #include "pub_core_vki.h"
 #include "pub_core_threadstate.h"
-#include "pub_core_debuginfo.h"
+#include "pub_core_debuginfo.h"     // XXX: circular dependency
 #include "pub_core_aspacemgr.h"     // For VG_(is_addressable)()
 #include "pub_core_libcbase.h"
 #include "pub_core_libcassert.h"
@@ -44,11 +44,14 @@
 #include "pub_core_clientstate.h"   // VG_(client__dl_sysinfo_int80)
 #include "pub_core_trampoline.h"
 
+
 /*------------------------------------------------------------*/
-/*--- Exported functions.                                  ---*/
+/*---                                                      ---*/
+/*--- BEGIN platform-dependent unwinder worker functions   ---*/
+/*---                                                      ---*/
 /*------------------------------------------------------------*/
 
-/* Take a snapshot of the client's stack, putting the up to 'n_ips'
+/* Take a snapshot of the client's stack, putting up to 'max_n_ips'
    IPs into 'ips'.  In order to be thread-safe, we pass in the
    thread's IP SP, FP if that's meaningful, and LR if that's
    meaningful.  Returns number of IPs put in 'ips'.
@@ -57,23 +60,17 @@
    first parameter, else send zero.  This helps generate better stack
    traces on ppc64-linux and has no effect on other platforms.
 */
-UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
-                               /*OUT*/Addr* ips, UInt n_ips,
-                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
-                               Addr ip, Addr sp, Addr fp, Addr lr,
-                               Addr fp_min, Addr fp_max_orig )
-{
-#  if defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
-                               || defined(VGP_ppc32_aix5) \
-                               || defined(VGP_ppc64_aix5)
-   Bool  lr_is_first_RA = False;
-#  endif
-#  if defined(VGP_ppc64_linux) || defined(VGP_ppc64_aix5) \
-                               || defined(VGP_ppc32_aix5)
-   Word redir_stack_size = 0;
-   Word redirs_used      = 0;
-#  endif
 
+/* ------------------------ x86 ------------------------- */
+
+#if defined(VGP_x86_linux) || defined(VGP_x86_darwin)
+
+UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
+                               /*OUT*/Addr* ips, UInt max_n_ips,
+                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
+                               UnwindStartRegs* startRegs,
+                               Addr fp_max_orig )
+{
    Bool  debug = False;
    Int   i;
    Addr  fp_max;
@@ -82,7 +79,13 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
    vg_assert(sizeof(Addr) == sizeof(UWord));
    vg_assert(sizeof(Addr) == sizeof(void*));
 
-   /* Snaffle IPs from the client's stack into ips[0 .. n_ips-1],
+   D3UnwindRegs uregs;
+   uregs.xip = (Addr)startRegs->r_pc;
+   uregs.xsp = (Addr)startRegs->r_sp;
+   uregs.xbp = startRegs->misc.X86.r_ebp;
+   Addr fp_min = uregs.xsp;
+
+   /* Snaffle IPs from the client's stack into ips[0 .. max_n_ips-1],
       stopping when the trail goes cold, which we guess to be
       when FP is not a reasonable stack location. */
 
@@ -94,36 +97,32 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
       fp_max -= sizeof(Addr);
 
    if (debug)
-      VG_(printf)("n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
+      VG_(printf)("max_n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
                   "fp_max=0x%lx ip=0x%lx fp=0x%lx\n",
-		  n_ips, fp_min, fp_max_orig, fp_max, ip, fp);
+                  max_n_ips, fp_min, fp_max_orig, fp_max,
+                  uregs.xip, uregs.xbp);
 
    /* Assertion broken before main() is reached in pthreaded programs;  the
     * offending stack traces only have one item.  --njn, 2002-aug-16 */
    /* vg_assert(fp_min <= fp_max);*/
+   // On Darwin, this kicks in for pthread-related stack traces, so they're
+   // only 1 entry long which is wrong.
+#  if !defined(VGO_darwin)
    if (fp_min + 512 >= fp_max) {
       /* If the stack limits look bogus, don't poke around ... but
          don't bomb out either. */
-      if (sps) sps[0] = sp;
-      if (fps) fps[0] = fp;
-      ips[0] = ip;
+      if (sps) sps[0] = uregs.xsp;
+      if (fps) fps[0] = uregs.xbp;
+      ips[0] = uregs.xip;
       return 1;
    } 
-
-   /* Otherwise unwind the stack in a platform-specific way.  Trying
-      to merge the x86, amd64, ppc32 and ppc64 logic into a single
-      piece of code is just too confusing and difficult to
-      performance-tune.  */
-
-#  if defined(VGP_x86_linux)
-
-   /*--------------------- x86 ---------------------*/
+#  endif
 
    /* fp is %ebp.  sp is %esp.  ip is %eip. */
 
-   if (sps) sps[0] = sp;
-   if (fps) fps[0] = fp;
-   ips[0] = ip;
+   if (sps) sps[0] = uregs.xsp;
+   if (fps) fps[0] = uregs.xbp;
+   ips[0] = uregs.xip;
    i = 1;
 
    /* Loop unwinding the stack. Note that the IP value we get on
@@ -143,7 +142,7 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
     */
    while (True) {
 
-      if (i >= n_ips)
+      if (i >= max_n_ips)
          break;
 
       /* Try to derive a new (ip,sp,fp) triple from the current
@@ -156,31 +155,55 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
          fails, and is expensive. */
       /* Deal with frames resulting from functions which begin "pushl%
          ebp ; movl %esp, %ebp" which is the ABI-mandated preamble. */
-      if (fp_min <= fp && fp <= fp_max
-                                - 1 * sizeof(UWord)/*see comment below*/) {
+      if (fp_min <= uregs.xbp &&
+          uregs.xbp <= fp_max - 1 * sizeof(UWord)/*see comment below*/)
+      {
          /* fp looks sane, so use it. */
-         ip = (((UWord*)fp)[1]);
-         sp = fp + sizeof(Addr) /*saved %ebp*/ 
-                 + sizeof(Addr) /*ra*/;
-         fp = (((UWord*)fp)[0]);
-         if (sps) sps[i] = sp;
-         if (fps) fps[i] = fp;
-         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
+         uregs.xip = (((UWord*)uregs.xbp)[1]);
+         // We stop if we hit a zero (the traditional end-of-stack
+         // marker) or a one -- these correspond to recorded IPs of 0 or -1.
+         // The latter because r8818 (in this file) changes the meaning of
+         // entries [1] and above in a stack trace, by subtracting 1 from
+         // them.  Hence stacks that used to end with a zero value now end in
+         // -1 and so we must detect that too.
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         uregs.xsp = uregs.xbp + sizeof(Addr) /*saved %ebp*/ 
+                               + sizeof(Addr) /*ra*/;
+         uregs.xbp = (((UWord*)uregs.xbp)[0]);
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
             VG_(printf)("     ipsF[%d]=0x%08lx\n", i-1, ips[i-1]);
-         ip = ip - 1; /* as per comment at the head of this loop */
+         uregs.xip = uregs.xip - 1;
+            /* as per comment at the head of this loop */
          continue;
       }
 
       /* That didn't work out, so see if there is any CF info to hand
          which can be used. */
-      if ( VG_(use_CF_info)( &ip, &sp, &fp, fp_min, fp_max ) ) {
-         if (sps) sps[i] = sp;
-         if (fps) fps[i] = fp;
-         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
+      if ( VG_(use_CF_info)( &uregs, fp_min, fp_max ) ) {
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
             VG_(printf)("     ipsC[%d]=0x%08lx\n", i-1, ips[i-1]);
-         ip = ip - 1; /* as per comment at the head of this loop */
+         uregs.xip = uregs.xip - 1;
+            /* as per comment at the head of this loop */
+         continue;
+      }
+
+      /* And, similarly, try for MSVC FPO unwind info. */
+      if ( VG_(use_FPO_info)( &uregs.xip, &uregs.xsp, &uregs.xbp,
+                              fp_min, fp_max ) ) {
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip;
+         if (debug)
+            VG_(printf)("     ipsC[%d]=0x%08lx\n", i-1, ips[i-1]);
+         uregs.xip = uregs.xip - 1;
          continue;
       }
 
@@ -188,15 +211,74 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
       break;
    }
 
-#  elif defined(VGP_amd64_linux)
+   n_found = i;
+   return n_found;
+}
 
-   /*--------------------- amd64 ---------------------*/
+#endif
+
+/* ----------------------- amd64 ------------------------ */
+
+#if defined(VGP_amd64_linux) || defined(VGP_amd64_darwin)
+
+UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
+                               /*OUT*/Addr* ips, UInt max_n_ips,
+                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
+                               UnwindStartRegs* startRegs,
+                               Addr fp_max_orig )
+{
+   Bool  debug = False;
+   Int   i;
+   Addr  fp_max;
+   UInt  n_found = 0;
+
+   vg_assert(sizeof(Addr) == sizeof(UWord));
+   vg_assert(sizeof(Addr) == sizeof(void*));
+
+   D3UnwindRegs uregs;
+   uregs.xip = startRegs->r_pc;
+   uregs.xsp = startRegs->r_sp;
+   uregs.xbp = startRegs->misc.AMD64.r_rbp;
+   Addr fp_min = uregs.xsp;
+
+   /* Snaffle IPs from the client's stack into ips[0 .. max_n_ips-1],
+      stopping when the trail goes cold, which we guess to be
+      when FP is not a reasonable stack location. */
+
+   // JRS 2002-sep-17: hack, to round up fp_max to the end of the
+   // current page, at least.  Dunno if it helps.
+   // NJN 2002-sep-17: seems to -- stack traces look like 1.0.X again
+   fp_max = VG_PGROUNDUP(fp_max_orig);
+   if (fp_max >= sizeof(Addr))
+      fp_max -= sizeof(Addr);
+
+   if (debug)
+      VG_(printf)("max_n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
+                  "fp_max=0x%lx ip=0x%lx fp=0x%lx\n",
+                  max_n_ips, fp_min, fp_max_orig, fp_max,
+                  uregs.xip, uregs.xbp);
+
+   /* Assertion broken before main() is reached in pthreaded programs;  the
+    * offending stack traces only have one item.  --njn, 2002-aug-16 */
+   /* vg_assert(fp_min <= fp_max);*/
+   // On Darwin, this kicks in for pthread-related stack traces, so they're
+   // only 1 entry long which is wrong.
+#  if !defined(VGO_darwin)
+   if (fp_min + 256 >= fp_max) {
+      /* If the stack limits look bogus, don't poke around ... but
+         don't bomb out either. */
+      if (sps) sps[0] = uregs.xsp;
+      if (fps) fps[0] = uregs.xbp;
+      ips[0] = uregs.xip;
+      return 1;
+   } 
+#  endif
 
    /* fp is %rbp.  sp is %rsp.  ip is %rip. */
 
-   ips[0] = ip;
-   if (sps) sps[0] = sp;
-   if (fps) fps[0] = fp;
+   ips[0] = uregs.xip;
+   if (sps) sps[0] = uregs.xsp;
+   if (fps) fps[0] = uregs.xbp;
    i = 1;
 
    /* Loop unwinding the stack. Note that the IP value we get on
@@ -216,21 +298,21 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
     */
    while (True) {
 
-      if (i >= n_ips)
+      if (i >= max_n_ips)
          break;
 
-      /* Try to derive a new (ip,sp,fp) triple from the current
-         set. */
+      /* Try to derive a new (ip,sp,fp) triple from the current set. */
 
       /* First off, see if there is any CFI info to hand which can
          be used. */
-      if ( VG_(use_CF_info)( &ip, &sp, &fp, fp_min, fp_max ) ) {
-         if (sps) sps[i] = sp;
-         if (fps) fps[i] = fp;
-         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
+      if ( VG_(use_CF_info)( &uregs, fp_min, fp_max ) ) {
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
             VG_(printf)("     ipsC[%d]=%#08lx\n", i-1, ips[i-1]);
-         ip = ip - 1; /* as per comment at the head of this loop */
+         uregs.xip = uregs.xip - 1; /* as per comment at the head of this loop */
          continue;
       }
 
@@ -246,18 +328,19 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
          fact that we are prodding at & ((UWord*)fp)[1] and so need to
          adjust the limit check accordingly.  Omitting this has been
          observed to cause segfaults on rare occasions. */
-      if (fp_min <= fp && fp <= fp_max - 1 * sizeof(UWord)) {
+      if (fp_min <= uregs.xbp && uregs.xbp <= fp_max - 1 * sizeof(UWord)) {
          /* fp looks sane, so use it. */
-         ip = (((UWord*)fp)[1]);
-         sp = fp + sizeof(Addr) /*saved %rbp*/ 
-                 + sizeof(Addr) /*ra*/;
-         fp = (((UWord*)fp)[0]);
-         if (sps) sps[i] = sp;
-         if (fps) fps[i] = fp;
-         ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
+         uregs.xip = (((UWord*)uregs.xbp)[1]);
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         uregs.xsp = uregs.xbp + sizeof(Addr) /*saved %rbp*/ 
+                               + sizeof(Addr) /*ra*/;
+         uregs.xbp = (((UWord*)uregs.xbp)[0]);
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip - 1; /* -1: refer to calling insn, not the RA */
          if (debug)
             VG_(printf)("     ipsF[%d]=%#08lx\n", i-1, ips[i-1]);
-         ip = ip - 1; /* as per comment at the head of this loop */
+         uregs.xip = uregs.xip - 1; /* as per comment at the head of this loop */
          continue;
       }
 
@@ -273,18 +356,20 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
            address; instead scan a likely section of stack (eg sp .. sp+256)
            and use suitable values found there.
       */
-      if (fp_min <= sp && sp < fp_max) {
-         ip = ((UWord*)sp)[0];
-         if (sps) sps[i] = sp;
-         if (fps) fps[i] = fp;
-         ips[i++] = ip == 0 
+      if (fp_min <= uregs.xsp && uregs.xsp < fp_max) {
+         uregs.xip = ((UWord*)uregs.xsp)[0];
+         if (0 == uregs.xip || 1 == uregs.xip) break;
+         if (sps) sps[i] = uregs.xsp;
+         if (fps) fps[i] = uregs.xbp;
+         ips[i++] = uregs.xip == 0 
                     ? 0 /* sp[0] == 0 ==> stuck at the bottom of a
                            thread stack */
-                    : ip - 1; /* -1: refer to calling insn, not the RA */
+                    : uregs.xip - 1;
+                        /* -1: refer to calling insn, not the RA */
          if (debug)
             VG_(printf)("     ipsH[%d]=%#08lx\n", i-1, ips[i-1]);
-         ip = ip - 1; /* as per comment at the head of this loop */
-         sp += 8;
+         uregs.xip = uregs.xip - 1; /* as per comment at the head of this loop */
+         uregs.xsp += 8;
          continue;
       }
 
@@ -292,10 +377,74 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
       break;
    }
 
-#  elif defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
-        || defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+   n_found = i;
+   return n_found;
+}
 
-   /*--------------------- ppc32/64 ---------------------*/
+#endif
+
+/* -----------------------ppc32/64 ---------------------- */
+
+#if defined(VGP_ppc32_linux) || defined(VGP_ppc64_linux) \
+    || defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+
+UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
+                               /*OUT*/Addr* ips, UInt max_n_ips,
+                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
+                               UnwindStartRegs* startRegs,
+                               Addr fp_max_orig )
+{
+   Bool  lr_is_first_RA = False;
+#  if defined(VG_PLAT_USES_PPCTOC)
+   Word redir_stack_size = 0;
+   Word redirs_used      = 0;
+#  endif
+
+   Bool  debug = False;
+   Int   i;
+   Addr  fp_max;
+   UInt  n_found = 0;
+
+   vg_assert(sizeof(Addr) == sizeof(UWord));
+   vg_assert(sizeof(Addr) == sizeof(void*));
+
+   Addr ip = (Addr)startRegs->r_pc;
+   Addr sp = (Addr)startRegs->r_sp;
+   Addr fp = sp;
+#  if defined(VGP_ppc32_linux) || defined(VGP_ppc32_aix5)
+   Addr lr = startRegs->misc.PPC32.r_lr;
+#  elif defined(VGP_ppc64_linux) || defined(VGP_ppc64_aix5)
+   Addr lr = startRegs->misc.PPC64.r_lr;
+#  endif
+   Addr fp_min = sp;
+
+   /* Snaffle IPs from the client's stack into ips[0 .. max_n_ips-1],
+      stopping when the trail goes cold, which we guess to be
+      when FP is not a reasonable stack location. */
+
+   // JRS 2002-sep-17: hack, to round up fp_max to the end of the
+   // current page, at least.  Dunno if it helps.
+   // NJN 2002-sep-17: seems to -- stack traces look like 1.0.X again
+   fp_max = VG_PGROUNDUP(fp_max_orig);
+   if (fp_max >= sizeof(Addr))
+      fp_max -= sizeof(Addr);
+
+   if (debug)
+      VG_(printf)("max_n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
+                  "fp_max=0x%lx ip=0x%lx fp=0x%lx\n",
+		  max_n_ips, fp_min, fp_max_orig, fp_max, ip, fp);
+
+   /* Assertion broken before main() is reached in pthreaded programs;  the
+    * offending stack traces only have one item.  --njn, 2002-aug-16 */
+   /* vg_assert(fp_min <= fp_max);*/
+   if (fp_min + 512 >= fp_max) {
+      /* If the stack limits look bogus, don't poke around ... but
+         don't bomb out either. */
+      if (sps) sps[0] = sp;
+      if (fps) fps[0] = fp;
+      ips[0] = ip;
+      return 1;
+   } 
 
    /* fp is %r1.  ip is %cia.  Note, ppc uses r1 as both the stack and
       frame pointers. */
@@ -337,8 +486,8 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
       UChar buf_lr[M_VG_ERRTXT], buf_ip[M_VG_ERRTXT];
       /* The following conditional looks grossly inefficient and
          surely could be majorly improved, with not much effort. */
-      if (VG_(get_fnname_nodemangle) (lr, buf_lr, M_VG_ERRTXT))
-         if (VG_(get_fnname_nodemangle) (ip, buf_ip, M_VG_ERRTXT))
+      if (VG_(get_fnname_raw) (lr, buf_lr, M_VG_ERRTXT))
+         if (VG_(get_fnname_raw) (ip, buf_ip, M_VG_ERRTXT))
             if (VG_(strncmp)(buf_lr, buf_ip, M_VG_ERRTXT))
                lr_is_first_RA = True;
 #     undef M_VG_ERRTXT
@@ -359,14 +508,13 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
         /* On ppc64-linux (ppc64-elf, really), and on AIX, the lr save
            slot is 2 words back from sp, whereas on ppc32-elf(?) it's
            only one word back. */
-#        if defined(VGP_ppc64_linux) \
-            || defined(VGP_ppc32_aix5) || defined(VGP_ppc64_aix5)
+#        if defined(VG_PLAT_USES_PPCTOC)
          const Int lr_offset = 2;
 #        else
          const Int lr_offset = 1;
 #        endif
 
-         if (i >= n_ips)
+         if (i >= max_n_ips)
             break;
 
          /* Try to derive a new (ip,fp) pair from the current set. */
@@ -402,9 +550,10 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
             }
 #           endif
 
-            fp = (((UWord*)fp)[0]);
+            if (0 == ip || 1 == ip) break;
             if (sps) sps[i] = fp; /* NB. not sp */
             if (fps) fps[i] = fp;
+            fp = (((UWord*)fp)[0]);
             ips[i++] = ip - 1; /* -1: refer to calling insn, not the RA */
             if (debug)
                VG_(printf)("     ipsF[%d]=%#08lx\n", i-1, ips[i-1]);
@@ -419,25 +568,129 @@ UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
       }
    }
 
-#  else
-#    error "Unknown platform"
-#  endif
+   n_found = i;
+   return n_found;
+}
+
+#endif
+
+/* ------------------------ arm ------------------------- */
+
+#if defined(VGP_arm_linux)
+
+UInt VG_(get_StackTrace_wrk) ( ThreadId tid_if_known,
+                               /*OUT*/Addr* ips, UInt max_n_ips,
+                               /*OUT*/Addr* sps, /*OUT*/Addr* fps,
+                               UnwindStartRegs* startRegs,
+                               Addr fp_max_orig )
+{
+   Bool  debug = False;
+   Int   i;
+   Addr  fp_max;
+   UInt  n_found = 0;
+
+   vg_assert(sizeof(Addr) == sizeof(UWord));
+   vg_assert(sizeof(Addr) == sizeof(void*));
+
+   D3UnwindRegs uregs;
+   uregs.r15 = startRegs->r_pc & 0xFFFFFFFE;
+   uregs.r14 = startRegs->misc.ARM.r14;
+   uregs.r13 = startRegs->r_sp;
+   uregs.r12 = startRegs->misc.ARM.r12;
+   uregs.r11 = startRegs->misc.ARM.r11;
+   uregs.r7  = startRegs->misc.ARM.r7;
+   Addr fp_min = uregs.r13;
+
+   /* Snaffle IPs from the client's stack into ips[0 .. max_n_ips-1],
+      stopping when the trail goes cold, which we guess to be
+      when FP is not a reasonable stack location. */
+
+   // JRS 2002-sep-17: hack, to round up fp_max to the end of the
+   // current page, at least.  Dunno if it helps.
+   // NJN 2002-sep-17: seems to -- stack traces look like 1.0.X again
+   fp_max = VG_PGROUNDUP(fp_max_orig);
+   if (fp_max >= sizeof(Addr))
+      fp_max -= sizeof(Addr);
+
+   if (debug)
+      VG_(printf)("\nmax_n_ips=%d fp_min=0x%lx fp_max_orig=0x%lx, "
+                  "fp_max=0x%lx r15=0x%lx r13=0x%lx\n",
+                  max_n_ips, fp_min, fp_max_orig, fp_max,
+                  uregs.r15, uregs.r13);
+
+   /* Assertion broken before main() is reached in pthreaded programs;  the
+    * offending stack traces only have one item.  --njn, 2002-aug-16 */
+   /* vg_assert(fp_min <= fp_max);*/
+   // On Darwin, this kicks in for pthread-related stack traces, so they're
+   // only 1 entry long which is wrong.
+   if (fp_min + 512 >= fp_max) {
+      /* If the stack limits look bogus, don't poke around ... but
+         don't bomb out either. */
+      if (sps) sps[0] = uregs.r13;
+      if (fps) fps[0] = 0;
+      ips[0] = uregs.r15;
+      return 1;
+   } 
+
+   /* */
+
+   if (sps) sps[0] = uregs.r13;
+   if (fps) fps[0] = 0;
+   ips[0] = uregs.r15;
+   i = 1;
+
+   /* Loop unwinding the stack. */
+
+   while (True) {
+      if (debug) {
+         VG_(printf)("i: %d, r15: 0x%lx, r13: 0x%lx\n",
+                     i, uregs.r15, uregs.r13);
+      }
+
+      if (i >= max_n_ips)
+         break;
+
+      if (VG_(use_CF_info)( &uregs, fp_min, fp_max )) {
+         if (sps) sps[i] = uregs.r13;
+         if (fps) fps[i] = 0;
+         ips[i++] = (uregs.r15 & 0xFFFFFFFE) - 1;
+         if (debug)
+            VG_(printf)("USING CFI: r15: 0x%lx, r13: 0x%lx\n",
+                        uregs.r15, uregs.r13);
+         uregs.r15 = (uregs.r15 & 0xFFFFFFFE) - 1;
+         continue;
+      }
+      /* No luck.  We have to give up. */
+      break;
+   }
 
    n_found = i;
    return n_found;
 }
 
+#endif
+
+/*------------------------------------------------------------*/
+/*---                                                      ---*/
+/*--- END platform-dependent unwinder worker functions     ---*/
+/*---                                                      ---*/
+/*------------------------------------------------------------*/
+
+/*------------------------------------------------------------*/
+/*--- Exported functions.                                  ---*/
+/*------------------------------------------------------------*/
+
 UInt VG_(get_StackTrace) ( ThreadId tid, 
-                           /*OUT*/StackTrace ips, UInt n_ips,
+                           /*OUT*/StackTrace ips, UInt max_n_ips,
                            /*OUT*/StackTrace sps,
                            /*OUT*/StackTrace fps,
                            Word first_ip_delta )
 {
-   /* thread in thread table */
-   Addr ip                 = VG_(get_IP)(tid);
-   Addr fp                 = VG_(get_FP)(tid);
-   Addr sp                 = VG_(get_SP)(tid);
-   Addr lr                 = VG_(get_LR)(tid);
+   /* Get the register values with which to start the unwind. */
+   UnwindStartRegs startRegs;
+   VG_(memset)( &startRegs, 0, sizeof(startRegs) );
+   VG_(get_UnwindStartRegs)( &startRegs, tid );
+
    Addr stack_highest_word = VG_(threads)[tid].client_stack_highest_word;
    Addr stack_lowest_word  = 0;
 
@@ -460,33 +713,35 @@ UInt VG_(get_StackTrace) ( ThreadId tid,
       bothered.
    */
    if (VG_(client__dl_sysinfo_int80) != 0 /* we know its address */
-       && ip >= VG_(client__dl_sysinfo_int80)
-       && ip < VG_(client__dl_sysinfo_int80)+3
-       && VG_(am_is_valid_for_client)(sp, sizeof(Addr), VKI_PROT_READ)) {
-      ip = *(Addr *)sp;
-      sp += sizeof(Addr);
+       && startRegs.r_pc >= VG_(client__dl_sysinfo_int80)
+       && startRegs.r_pc < VG_(client__dl_sysinfo_int80)+3
+       && VG_(am_is_valid_for_client)(startRegs.r_pc, sizeof(Addr),
+                                      VKI_PROT_READ)) {
+      startRegs.r_pc  = (ULong) *(Addr*)(UWord)startRegs.r_sp;
+      startRegs.r_sp += (ULong) sizeof(Addr);
    }
 #  endif
 
    /* See if we can get a better idea of the stack limits */
-   VG_(stack_limits)(sp, &stack_lowest_word, &stack_highest_word);
+   VG_(stack_limits)( (Addr)startRegs.r_sp,
+                      &stack_lowest_word, &stack_highest_word );
 
    /* Take into account the first_ip_delta. */
-   vg_assert( sizeof(Addr) == sizeof(Word) );
-   ip += first_ip_delta;
+   startRegs.r_pc += (Long)(Word)first_ip_delta;
 
    if (0)
-      VG_(printf)("tid %d: stack_highest=0x%08lx ip=0x%08lx "
-                  "sp=0x%08lx fp=0x%08lx\n",
-		  tid, stack_highest_word, ip, sp, fp);
+      VG_(printf)("tid %d: stack_highest=0x%08lx ip=0x%010llx "
+                  "sp=0x%010llx\n",
+		  tid, stack_highest_word,
+                  startRegs.r_pc, startRegs.r_sp);
 
-   return VG_(get_StackTrace_wrk)(tid, ips, n_ips, 
+   return VG_(get_StackTrace_wrk)(tid, ips, max_n_ips, 
                                        sps, fps,
-                                       ip, sp, fp, lr, sp, 
+                                       &startRegs,
                                        stack_highest_word);
 }
 
-static void printIpDesc(UInt n, Addr ip)
+static void printIpDesc(UInt n, Addr ip, void* uu_opaque)
 {
    #define BUF_LEN   4096
    
@@ -495,9 +750,9 @@ static void printIpDesc(UInt n, Addr ip)
    VG_(describe_IP)(ip, buf, BUF_LEN);
 
    if (VG_(clo_xml)) {
-      VG_(message)(Vg_UserMsg, "    %s", buf);
+      VG_(printf_xml)("    %s\n", buf);
    } else {
-      VG_(message)(Vg_UserMsg, "   %s %s", ( n == 0 ? "at" : "by" ), buf);
+      VG_(message)(Vg_UserMsg, "   %s %s\n", ( n == 0 ? "at" : "by" ), buf);
    }
 }
 
@@ -507,35 +762,33 @@ void VG_(pp_StackTrace) ( StackTrace ips, UInt n_ips )
    vg_assert( n_ips > 0 );
 
    if (VG_(clo_xml))
-      VG_(message)(Vg_UserMsg, "  <stack>");
+      VG_(printf_xml)("  <stack>\n");
 
-   VG_(apply_StackTrace)( printIpDesc, ips, n_ips );
+   VG_(apply_StackTrace)( printIpDesc, NULL, ips, n_ips );
 
    if (VG_(clo_xml))
-      VG_(message)(Vg_UserMsg, "  </stack>");
+      VG_(printf_xml)("  </stack>\n");
 }
 
 /* Get and immediately print a StackTrace. */
-void VG_(get_and_pp_StackTrace) ( ThreadId tid, UInt n_ips )
+void VG_(get_and_pp_StackTrace) ( ThreadId tid, UInt max_n_ips )
 {
-   Addr ips[n_ips];
-   UInt n_ips_obtained 
-      = VG_(get_StackTrace)(tid, ips, n_ips,
+   Addr ips[max_n_ips];
+   UInt n_ips
+      = VG_(get_StackTrace)(tid, ips, max_n_ips,
                             NULL/*array to dump SP values in*/,
                             NULL/*array to dump FP values in*/,
                             0/*first_ip_delta*/);
-   VG_(pp_StackTrace)(ips, n_ips_obtained);
+   VG_(pp_StackTrace)(ips, n_ips);
 }
 
-
-void VG_(apply_StackTrace)( void(*action)(UInt n, Addr ip),
-                            StackTrace ips, UInt n_ips )
+void VG_(apply_StackTrace)(
+        void(*action)(UInt n, Addr ip, void* opaque),
+        void* opaque,
+        StackTrace ips, UInt n_ips
+     )
 {
-   #define MYBUF_LEN 50  // only needs to be long enough for 
-                         // the names specially tested for
-
    Bool main_done = False;
-   Char mybuf[MYBUF_LEN];     // ok to stack allocate mybuf[] -- it's tiny
    Int i = 0;
 
    vg_assert(n_ips > 0);
@@ -545,29 +798,18 @@ void VG_(apply_StackTrace)( void(*action)(UInt n, Addr ip),
       // Stop after the first appearance of "main" or one of the other names
       // (the appearance of which is a pretty good sign that we've gone past
       // main without seeing it, for whatever reason)
-      if ( ! VG_(clo_show_below_main)) {
-         VG_(get_fnname_nodemangle)( ip, mybuf, MYBUF_LEN );
-         mybuf[MYBUF_LEN-1] = 0; // paranoia
-         if ( VG_STREQ("main", mybuf)
-#             if defined(VGO_linux)
-              || VG_STREQ("__libc_start_main", mybuf)   // glibc glibness
-              || VG_STREQ("generic_start_main", mybuf)  // Yellow Dog doggedness
-#             endif
-            )
+      if ( ! VG_(clo_show_below_main) ) {
+         Vg_FnNameKind kind = VG_(get_fnname_kind_from_IP)(ip);
+         if (Vg_FnNameMain == kind || Vg_FnNameBelowMain == kind) {
             main_done = True;
+         }
       }
 
       // Act on the ip
-      action(i, ip);
+      action(i, ip, opaque);
 
       i++;
-      // re 'while' condition: stop if we hit a zero value (the traditional
-      // end-of-stack marker) or a ~0 value.  The latter because r8818 
-      // (in this file) changes the meaning of entries [1] and above in a
-      // stack trace, by subtracting 1 from them.  Hence stacks that used
-      // to end with a zero value now end in -1 and so we must detect
-      // that too.
-   } while (i < n_ips && ips[i] != 0 && ips[i] != ~(Addr)0 && !main_done);
+   } while (i < n_ips && !main_done);
 
    #undef MYBUF_LEN
 }
