@@ -7,7 +7,7 @@
    This file is part of Valgrind, a dynamic binary instrumentation
    framework.
 
-   Copyright (C) 2000-2008 Julian Seward
+   Copyright (C) 2000-2010 Julian Seward
       jseward@acm.org
 
    This program is free software; you can redistribute it and/or
@@ -59,6 +59,11 @@ extern void VG_(nuke_all_threads_except) ( ThreadId me,
    thread. */
 extern void VG_(acquire_BigLock) ( ThreadId tid, HChar* who );
 
+/* Simple version, which simply acquires the lock, but does not mess
+   with the guest state in the same way as the non _LL version
+   does. */
+extern void VG_(acquire_BigLock_LL) ( HChar* who );
+
 /* Set a thread into a sleeping state.  Before the call, the thread
    must be runnable, and holding the CPU lock.  When this call
    returns, the thread will be set to the specified sleeping state,
@@ -67,9 +72,14 @@ extern void VG_(acquire_BigLock) ( ThreadId tid, HChar* who );
    caller must be careful not to touch any shared state.  It is also
    the caller's responsibility to actually block until the thread is
    ready to run again. */
-extern void VG_(release_BigLock) ( ThreadId tid, ThreadStatus state, HChar* who );
+extern void VG_(release_BigLock) ( ThreadId tid,
+                                   ThreadStatus state, HChar* who );
 
-/* Yield the CPU for a while */
+/* Matching function to acquire_BigLock_LL. */
+extern void VG_(release_BigLock_LL) ( HChar* who );
+
+/* Yield the CPU for a while.  Drops/acquires the lock using the
+   normal (non _LL) functions. */
 extern void VG_(vg_yield)(void);
 
 // The scheduler.
