@@ -2,7 +2,7 @@
 /*
   This file is part of drd, a thread error detector.
 
-  Copyright (C) 2006-2009 Bart Van Assche <bart.vanassche@gmail.com>.
+  Copyright (C) 2006-2010 Bart Van Assche <bvanassche@acm.org>.
 
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
@@ -190,7 +190,7 @@ void drd_report_data_race(Error* const err, const DataRaceErrInfo* const dri)
 /**
  * Compare two error contexts. The core function VG_(maybe_record_error)()
  * calls this function to compare error contexts such that errors that occur
- * repeatedly are only printed once. This function is only called by the core 
+ * repeatedly are only printed once. This function is only called by the core
  * if the error kind of e1 and e2 matches and if the ExeContext's of e1 and
  * e2 also match.
  */
@@ -371,9 +371,11 @@ static void drd_tool_error_pp(Error* const e)
       break;
    }
    case GenericErr: {
-      //GenericErrInfo* gei =(GenericErrInfo*)(VG_(get_error_extra)(e));
+      GenericErrInfo* gei = (GenericErrInfo*)(VG_(get_error_extra)(e));
       VG_(message)(Vg_UserMsg, "%s\n", VG_(get_error_string)(e));
       VG_(pp_ExeContext)(VG_(get_error_where)(e));
+      if (gei->addr)
+	 first_observed(gei->addr);
       break;
    }
    case InvalidThreadId: {
@@ -383,15 +385,24 @@ static void drd_tool_error_pp(Error* const e)
       VG_(pp_ExeContext)(VG_(get_error_where)(e));
       break;
    }
-   case UnimpClReq: {
+   case UnimpHgClReq: {
       UnimpClReqInfo* uicr =(UnimpClReqInfo*)(VG_(get_error_extra)(e));
       VG_(message)(Vg_UserMsg,
                    "The annotation macro %s has not yet been implemented in"
-                   " <helgrind/helgrind.h>\n",
+                   " <valgrind/helgrind.h>\n",
                    /*VG_(get_error_string)(e),*/ uicr->descr);
       VG_(pp_ExeContext)(VG_(get_error_where)(e));
       break;
    }
+   case UnimpDrdClReq: {
+      UnimpClReqInfo* uicr =(UnimpClReqInfo*)(VG_(get_error_extra)(e));
+      VG_(message)(Vg_UserMsg,
+                   "The annotation macro %s has not yet been implemented in"
+                   " <valgrind/drd.h>\n",
+                   uicr->descr);
+      VG_(pp_ExeContext)(VG_(get_error_where)(e));
+      break;
+   }
    default:
       VG_(message)(Vg_UserMsg,
                    "%s\n",
@@ -429,7 +440,9 @@ static UInt drd_tool_error_update_extra(Error* e)
       return sizeof(GenericErrInfo);
    case InvalidThreadId:
       return sizeof(InvalidThreadIdInfo);
-   case UnimpClReq:
+   case UnimpHgClReq:
+      return sizeof(UnimpClReqInfo);
+   case UnimpDrdClReq:
       return sizeof(UnimpClReqInfo);
    default:
       tl_assert(False);
@@ -472,8 +485,10 @@ static Bool drd_is_recognized_suppression(Char* const name, Supp* const supp)
       skind = GenericErr;
    else if (VG_(strcmp)(name, STR_InvalidThreadId) == 0)
       skind = InvalidThreadId;
-   else if (VG_(strcmp)(name, STR_UnimpClReq) == 0)
-      skind = UnimpClReq;
+   else if (VG_(strcmp)(name, STR_UnimpHgClReq) == 0)
+      skind = UnimpHgClReq;
+   else if (VG_(strcmp)(name, STR_UnimpDrdClReq) == 0)
+      skind = UnimpDrdClReq;
    else
       return False;
 
@@ -520,7 +535,8 @@ static Char* drd_get_error_name(Error* e)
    case HoldtimeErr:  return VGAPPEND(STR_, HoldtimeErr);
    case GenericErr:   return VGAPPEND(STR_, GenericErr);
    case InvalidThreadId: return VGAPPEND(STR_, InvalidThreadId);
-   case UnimpClReq:   return VGAPPEND(STR_, UnimpClReq);
+   case UnimpHgClReq:  return VGAPPEND(STR_, UnimpHgClReq);
+   case UnimpDrdClReq: return VGAPPEND(STR_, UnimpDrdClReq);
    default:
       tl_assert(0);
    }
